/*
 * class_reference.h
 *
 *  Created on: Oct 28, 2010
 *      Author: philipp_
 */

#ifndef CLASS_REFERENCE_H_
#define CLASS_REFERENCE_H_

#include <iostream>
#include <vector>
#include <fstream>

#include "../parser/parameter.h"

using std::vector;
using std::pair;
using std::streampos;
using std::cout;
using std::endl;

class Reference {
private:
	streampos file_pos;
	string filename;
	//	vector<string> name;
	//	int scaff_pos;
	int prev_ID;
	size_t overlap;
	size_t mapsize;
	bool direction;

	void init() {

		direction = true;
		scaff_pos = 0;
		file_pos = 0;
	}

	void init_lookup() {
		clear_scaff_keys();

		for (size_t i = 0; i < mapsize; i++) {
			vector<pair<int, int> > temp;
			lookup_scaffold.push_back(temp);
		}
	}

	void clear_lookup() {

		lookup_scaffold.clear();
	}

public:
	int scaff_pos;
	bool flag;

	Reference(size_t p_mapsize, size_t over) {
		prev_ID = 0;
		flag = false;
		mapsize = p_mapsize;
		//init_lookup();
		init();
		overlap = over;
	}
	Reference(size_t over) {
		flag = false;
		mapsize=0;
		prev_ID = 0;
		init();
		overlap = over;
	}

	~Reference() {
		clear_scaff_keys();
		filename.clear();
		name.clear();

	}
	vector<string> name;
	//	size_t nucs;
	vector<string> sequence;
	vector<vector<pair<int, int> > > lookup_scaffold;

	//========== setter: ============
	void set_id(size_t id_new) {
		prev_ID = id_new;
	}

	void set_name(string new_name) {
		cout << "do not use" << endl;
		//exit(0);
		name[prev_ID] = new_name;
		scaff_pos = 0;
	}
	void set_file_pos(streampos pos) {
		file_pos = pos;
	}

	void set_scaff_pos(int pos) {
		scaff_pos = pos;
	}
	void set_filename(string file) {
		filename.clear();
		filename = file;
		init();
	}
	//	void set_filename_pair(string file){
	//		prev_ID=0;
	//		flag=false;
	//		for(int i =0; i<mapsize;i++){
	//			vector< pair<int,int> > temp;
	//			lookup_scaffold.push_back(temp);
	//		}
	//		init();
	//		overlap=over;
	//	}

	void set_overlap(int lap) {
		overlap = lap;
	}
	void set_direction(bool dir) {
		direction = dir;
	}
	//============== getter ==============

	streampos get_file_pos() {
		return file_pos;
	}

	string get_name(size_t id) {
		if (id < name.size()) {
			return name[id];
		}
		return "";
	}

	int get_ID() {
		return prev_ID;
	}

	int get_scaff_pos(int id) {
		if (id == prev_ID) {
			return scaff_pos;
		}
		return 0;
	}

	const char * get_filename() {
		return filename.c_str();
	}

	size_t get_overlap() {
		return overlap;
	}
	bool get_direction() {
		return direction;
	}
	//========== functions: =================

	void clear_names(){
		name.clear();
	}

	void inc_scaff_pos(int inc) {
		set_direction(true);
		scaff_pos += inc;
	}

	//converts the objects sequence to the complement
	void to_complement() {
		set_direction(false);
		vector < string > sequence2;
		string seq;

		for (vector<string>::iterator j = sequence.begin(); j != sequence.end(); j++) {
			for (string::iterator i = (*j).begin(); i != (*j).end(); i++) {
				switch ((*i)) {
				case 'A':
					seq += 'T';
					break;
				case 'C':
					seq += 'G';
					break;
				case 'G':
					seq += 'C';
					break;
				case 'T':
					seq += 'A';
					break;
				default:
					seq += 'N';
					break;
				}
			}
			sequence2.push_back(seq);
			seq.clear();
		}

		sequence.clear();
		sequence = sequence2;
	}

	void map_key(Parameters parameters) {
		init_lookup();
		int key = 0;
		for (size_t id = 0; id < sequence.size(); id++) {
			for (size_t i = 0; i + parameters.wordsize < sequence[id].size(); i++) {
				key = FatsHash(sequence[id].substr(i, parameters.wordsize));//, parameters);
				if (key >= 0 && key <= parameters.mapsize) {//&& prev_key!= key){
					lookup_scaffold[key].push_back(pair<int, int> ((int)i, prev_ID +(int) id));
				}
			}
		}
	}

	void clear_scaff_keys() {
		for (size_t i = 0; i < lookup_scaffold.size(); ++i) {
			lookup_scaffold[i].clear();
		}
		clear_lookup();
	}

	void map_BS_key(Parameters parameters) {
		int key = 0;
		init_lookup();

		for (size_t id = 0; id < sequence.size(); id++) {
			for (size_t i = 0; i + parameters.wordsize < sequence[id].size(); i++) {
				key = FatsHash(sequence[id].substr(i, parameters.wordsize));//, parameters);
				if (key >= 0 && key <= parameters.mapsize) {//&& prev_key!= key){
					lookup_scaffold[key].push_back(pair<int, int> ((int) i, prev_ID + (int) id));
				}
			}
		}
	}

};

#endif /* CLASS_REFERENCE_H_ */
