/*
 * SW_CUDA_Function.cu
 *
 *  Created on: Nov 23, 2009
 *      Author: fritz
 */

#include "cuda_sw.cuh"
#include <list>
#include <map>
using namespace std;

#include "functions/cuda_functions.cuh"

class is_near_str {
	int diff;

public:
	is_near_str(int val) {
		diff = val;
	}
	bool operator()(ids_str &first, ids_str &second) {
		if (abs(first.position - second.position) <= diff) {
			first.frequ++;
			return true;
		}
		return false;

	}
};

class is_near_pair_str {
	int diff;

public:
	is_near_pair_str(int val) {
		diff = val;
	}
	bool operator()(pair<int, int> &first, pair<int, int> &second) {
		if (abs(first.first - second.first) <= diff) {
			first.second++;
			return true;
		}
		return false;

	}
};

bool compare_pair_str(pair<int, int> first, pair<int, int> second) {
	return (first.first < second.first);
}

//computes the possible positions per read and prepares the sequences
int pre_filter_SW(map<int, list<ids_str> > &liste, iterators_str & iter,
		char * & reads_cpu, char * & scaff_cpu, vector<HelpStruct> & helper1,
		Reference & scaffold, Parameters parameters, bool & finished) {

	int t1 = 0;
	if (!finished) { // if there are still positions in the holder
		for (map<int, list<ids_str> >::iterator i = liste.begin(); i
				!= liste.end(); i++) {
			(*i).second.sort(compare_str); //sort according the position
			(*i).second.unique(is_near_str(iter.tresh)); //remove redundant keys and count their frequency given a position

			//run over non redundant list
			for (list<ids_str>::iterator it = (*i).second.begin(); it
					!= (*i).second.end(); ++it) {
				if ((*it).frequ >= parameters.seedword_hits) { //checks the frequency of keys given a position
					//copy read sequece
					memcpy(&reads_cpu[t1 * iter.seqlength],
							(*iter.read_pos).c_str(), iter.seqlength
									* sizeof(char));

					//copy reference seqeuence
					int pos = (*it).position - iter.shift;
					if (pos >= 0 && pos + iter.size
							< scaffold.sequence[(*it).scaff_id
									- scaffold.get_ID()].size()) {
						memcpy(
								&scaff_cpu[t1 * iter.size],
								scaffold.sequence[(*it).scaff_id
										- scaffold.get_ID()].substr(
										(size_t) pos, (size_t) iter.size).c_str(),
								iter.size * sizeof(char));
					} else {
						//if a read hangs over the begining or end of the reference
						size_t z = 0;
						for (int i = pos; i < pos + iter.size; i++) {
							if (i < 0 || i >= scaffold.sequence[(*it).scaff_id
									- scaffold.get_ID()].size()) {
								scaff_cpu[z + iter.size * t1] = 'N';
							} else {
								scaff_cpu[z + iter.size * t1]
										= scaffold.sequence[(*it).scaff_id
												- scaffold.get_ID()][i];
							}
							z++;
						}
					}

					//store meta information for post processing
					helper1[t1].id = iter.readID;
					helper1[t1].position = (*it).position;
					helper1[t1].scaff_id = (*it).scaff_id;

					t1++; //count the number of threads needed

					if (t1 == iter.proc) {//if number is reached, return.
						(*i).second.erase((*i).second.begin(), it);
						finished = false;
						return t1;
					}
				}
			}
			(*i).second.clear();
		}

		finished = true;

		for (map<int, list<ids_str> >::iterator i = liste.begin(); i
				!= liste.end(); i++) {
			(*i).second.clear();
		}
		liste.clear();
		iter.readID++;
		iter.mapps++;
		iter.match++;
		iter.read_pos++;
	}

	while (iter.readID < iter.numofSeq) {

		if (parameters.max_equal_scoring>1
				|| (*iter.match).match[parameters.X_best_hits - 1].score
						< (*iter.match).orig_length * parameters.match[0]) {
			ids_str temp;
			temp.frequ = 1;
			//collect positions
			//run over all keys per read
			for (vector<HashKey>::iterator i = (*iter.mapps).begin(); i	!= (*iter.mapps).end(); i++) {

				//get all positions of the key in the reference
				for (vector<pair<int, int> >::iterator j =scaffold.lookup_scaffold[(*i).key].begin(); j!= scaffold.lookup_scaffold[(*i).key].end(); j++) {
					temp.position = (*j).first - (*i).pos; //calc theoretically starting positon
					temp.scaff_id = (*j).second; //stores which reference (if there are several smaller ones)
					liste[(*j).second].push_back(temp);
				}
			}

			for (map<int, list<ids_str> >::iterator j = liste.begin(); j
					!= liste.end(); j++) {
				//filter and count frequencies:
				(*j).second.sort(compare_str);//sort according the start position
				(*j).second.unique(is_near_str(iter.tresh)); //delete redundant information and count the frequency of keys within a range

				//				if(parameters.speed){
				//					//delete all putative positions apart form the best 50
				//					(*j).second.sort(compare_matches_str);
				//
				//					list<ids_str>::iterator ttt =(*j).second.begin();
				//					int z=0; //neither advance nor just adding 50 to the iterate works ????
				//					while(z<51){
				//						z++;
				//						ttt++;
				//					}
				//					(*j).second.erase(ttt,(*j).second.end());
				//				}

				//runs over non redundant list and collects all sequences needed:
				for (list<ids_str>::iterator it = (*j).second.begin(); it
						!= (*j).second.end(); ++it) {
					if ((*it).frequ >= parameters.seedword_hits) { // if frequency of keys within a position is enough
						//coopy read sequence
						memcpy(&reads_cpu[t1 * iter.seqlength],
								(*iter.read_pos).c_str(), iter.seqlength
										* sizeof(char));

						//copy reference piece
						int pos = (*it).position - iter.shift;
						if (pos >= 0 && pos + iter.size
								< scaffold.sequence[(*it).scaff_id
										- scaffold.get_ID()].size()) {
							memcpy(
									&scaff_cpu[t1 * iter.size],
									scaffold.sequence[(*it).scaff_id
											- scaffold.get_ID()].substr(
											(size_t) pos, (size_t) iter.size).c_str(),
									iter.size * sizeof(char));
						} else {
							//if read overhang on either the begining or the end of the reference
							size_t z = 0;
							for (int i = pos; i < pos + iter.size; i++) {
								if (i < 0 || i
										>= scaffold.sequence[(*it).scaff_id
												- scaffold.get_ID()].size()) {
									scaff_cpu[z + iter.size * t1] = 'N';
								} else {
									scaff_cpu[z + iter.size * t1]
											= scaffold.sequence[(*it).scaff_id
													- scaffold.get_ID()][i];
								}
								z++;
							}
						}

						//stores meta informations:
						helper1[t1].id = iter.readID;
						helper1[t1].position = (*it).position;
						helper1[t1].scaff_id = (*it).scaff_id;

						t1++;
						if (t1 == iter.proc) { //if we collected enough data return
							(*j).second.erase((*j).second.begin(), it);
							finished = false; //the list is not finished
							return t1;
						}
					}
				}

				(*j).second.clear();
			}
			liste.clear();
		}
		finished = true;
		for (map<int, list<ids_str> >::iterator i = liste.begin(); i
				!= liste.end(); i++) {
			(*i).second.clear();
		}
		liste.clear();
		iter.readID++;
		iter.mapps++;
		iter.match++;
		iter.read_pos++;
	}

	memset(&reads_cpu[t1 * iter.seqlength], ' ', (iter.proc - t1)
			* iter.seqlength * sizeof(char));
	return t1;
}

int pre_filter_SW_BS(map<int, list<ids_str> > &liste, iterators_str & iter,
		char * & reads_cpu, char * & scaff_cpu, vector<HelpStruct> & helper1,
		Reference & scaffold, Parameters parameters, bool & finished, vector< vector <int > > bslookup) {

	int t1 = 0;
	if (!finished) { // if there are still positions in the holder
		for (map<int, list<ids_str> >::iterator i = liste.begin(); i
				!= liste.end(); i++) {
			(*i).second.sort(compare_str); //sort according the position
			(*i).second.unique(is_near_str(iter.tresh)); //remove redundant keys and count their frequency given a position

			//run over non redundant list
			for (list<ids_str>::iterator it = (*i).second.begin(); it
					!= (*i).second.end(); ++it) {
				if ((*it).frequ >= parameters.seedword_hits) { //checks the frequency of keys given a position
					//copy read sequece
					memcpy(&reads_cpu[t1 * iter.seqlength],	(*iter.read_pos).c_str(), iter.seqlength* sizeof(char));

					//copy reference seqeuence
					int pos = (*it).position - iter.shift;
					if (pos >= 0 && pos + iter.size< scaffold.sequence[(*it).scaff_id- scaffold.get_ID()].size()) {
						memcpy(&scaff_cpu[t1 * iter.size],scaffold.sequence[(*it).scaff_id- scaffold.get_ID()].substr((size_t) pos, (size_t) iter.size).c_str(),iter.size * sizeof(char));
					} else {
						//if a read hangs over the begining or end of the reference
						size_t z = 0;
						for (int i = pos; i < pos + iter.size; i++) {
							if (i < 0 || i >= scaffold.sequence[(*it).scaff_id
									- scaffold.get_ID()].size()) {
								scaff_cpu[z + iter.size * t1] = 'N';
							} else {
								scaff_cpu[z + iter.size * t1]
										= scaffold.sequence[(*it).scaff_id
												- scaffold.get_ID()][i];
							}
							z++;
						}
					}

					//store meta information for post processing
					helper1[t1].id = iter.readID;
					helper1[t1].position = (*it).position;
					helper1[t1].scaff_id = (*it).scaff_id;


					t1++; //count the number of threads needed

					if (t1 == iter.proc) {//if number is reached, return.
						(*i).second.erase((*i).second.begin(), it);
						finished = false;
						return t1;
					}
				}
			}
			(*i).second.clear();
		}

		finished = true;

		for (map<int, list<ids_str> >::iterator i = liste.begin(); i
				!= liste.end(); i++) {
			(*i).second.clear();
		}
		liste.clear();
		iter.readID++;
		iter.mapps++;
		iter.match++;
		iter.read_pos++;
	}

	while (iter.readID < iter.numofSeq) {

		if (parameters.max_equal_scoring>1
				|| (*iter.match).match[parameters.X_best_hits - 1].score
						< (*iter.match).orig_length * parameters.match[0]) {
			ids_str temp;
			temp.frequ = 1;
			//collect positions
			//run over all keys per read
			for (vector<HashKey>::iterator i = (*iter.mapps).begin(); i!= (*iter.mapps).end(); i++) {

				//get all positions of the key in the reference
				for (vector<pair<int, int> >::iterator j =	scaffold.lookup_scaffold[(*i).key].begin(); j!= scaffold.lookup_scaffold[(*i).key].end(); j++) {
					temp.position = (*j).first - (*i).pos; //calc theoretically starting positon
					temp.scaff_id = (*j).second; //stores which reference (if there are several smaller ones)
					liste[(*j).second].push_back(temp);
				}

				//get all position for every permutation in the reference
				for(vector<int>::iterator k= bslookup[(*i).key].begin(); k!=bslookup[(*i).key].end();k++){
					for (vector<pair<int, int> >::iterator j =	scaffold.lookup_scaffold[(*k)].begin(); j!= scaffold.lookup_scaffold[(*k)].end(); j++) {
						temp.position = (*j).first - (*i).pos; //calc theoretically starting positon
						temp.scaff_id = (*j).second; //stores which reference (if there are several smaller ones)
						liste[(*j).second].push_back(temp);
					}
				}
			}

			for (map<int, list<ids_str> >::iterator j = liste.begin(); j
					!= liste.end(); j++) {
				//filter and count frequencies:
				(*j).second.sort(compare_str);//sort according the start position
				(*j).second.unique(is_near_str(iter.tresh)); //delete redundant information and count the frequency of keys within a range

				//				if(parameters.speed){
				//					//delete all putative positions apart form the best 50
				//					(*j).second.sort(compare_matches_str);
				//
				//					list<ids_str>::iterator ttt =(*j).second.begin();
				//					int z=0; //neither advance nor just adding 50 to the iterate works ????
				//					while(z<51){
				//						z++;
				//						ttt++;
				//					}
				//					(*j).second.erase(ttt,(*j).second.end());
				//				}

				//runs over non redundant list and collects all sequences needed:
				for (list<ids_str>::iterator it = (*j).second.begin(); it
						!= (*j).second.end(); ++it) {
					if ((*it).frequ >= parameters.seedword_hits) { // if frequency of keys within a position is enough
						//coopy read sequence
						memcpy(&reads_cpu[t1 * iter.seqlength],
								(*iter.read_pos).c_str(), iter.seqlength
										* sizeof(char));

						//copy reference piece
						int pos = (*it).position - iter.shift;
						if (pos >= 0 && pos + iter.size
								< scaffold.sequence[(*it).scaff_id
										- scaffold.get_ID()].size()) {
							memcpy(
									&scaff_cpu[t1 * iter.size],
									scaffold.sequence[(*it).scaff_id
											- scaffold.get_ID()].substr(
											(size_t) pos, (size_t) iter.size).c_str(),
									iter.size * sizeof(char));
						} else {
							//if read overhang on either the begining or the end of the reference
							size_t z = 0;
							for (int i = pos; i < pos + iter.size; i++) {
								if (i < 0 || i
										>= scaffold.sequence[(*it).scaff_id
												- scaffold.get_ID()].size()) {
									scaff_cpu[z + iter.size * t1] = 'N';
								} else {
									scaff_cpu[z + iter.size * t1]
											= scaffold.sequence[(*it).scaff_id
													- scaffold.get_ID()][i];
								}
								z++;
							}
						}

						//stores meta informations:
						helper1[t1].id = iter.readID;
						helper1[t1].position = (*it).position;
						helper1[t1].scaff_id = (*it).scaff_id;

						t1++;
						if (t1 == iter.proc) { //if we collected enough data return
							(*j).second.erase((*j).second.begin(), it);
							finished = false; //the list is not finished
							return t1;
						}
					}
				}

				(*j).second.clear();
			}
			liste.clear();
		}
		finished = true;
		for (map<int, list<ids_str> >::iterator i = liste.begin(); i
				!= liste.end(); i++) {
			(*i).second.clear();
		}
		liste.clear();
		iter.readID++;
		iter.mapps++;
		iter.match++;
		iter.read_pos++;
	}

	memset(&reads_cpu[t1 * iter.seqlength], ' ', (iter.proc - t1)
			* iter.seqlength * sizeof(char));
	return t1;
}
//=============================================================================================


//process the results of the SW
void post_proc_SW(float * results_cpu, Reads & reads, Reference & ref,
		float *& results_gpu, Parameters parameters, int & t2, long mem_result,
		vector<HelpStruct> helper2, bool rev) {

	//copy results form GPU
	CUDA_SAFE_CALL(cudaMemcpy(results_cpu, results_gpu, mem_result
			* sizeof(float), cudaMemcpyDeviceToHost));

	//runs over the results
	for (size_t k = 0; k < t2; k++) {
		if (parameters.X_best_hits == 1 && results_cpu[k]
				>= reads.matches[helper2[k].id].match[0].score) { //TODO: nbest positions
			//new best score
			if (results_cpu[k] > reads.matches[helper2[k].id].match[0].score) {
				store(reads.matches[helper2[k].id], ref, helper2[k].position,
						helper2[k].scaff_id, results_cpu[k],
						reads.get_direction(), rev);
			} else {
				//equal scoring position
				if (reads.matches[helper2[k].id].match.size()
						< parameters.max_equal_scoring) {
					insert_equal_scoring(reads.matches[helper2[k].id], ref,
							helper2[k].position, helper2[k].scaff_id,
							results_cpu[k], reads.get_direction(),
							parameters.corridor_with, rev);
				}
			}
		} else if (parameters.X_best_hits > 1) {
			//store no optimal scores and positions!
			size_t size = reads.matches[helper2[k].id].match.size();
			if (reads.matches[helper2[k].id].match[size - 1].score
					< results_cpu[k]) {
				//store
				insert_sort(reads.matches[helper2[k].id], ref,
						helper2[k].position, helper2[k].scaff_id,
						results_cpu[k], reads.get_direction(),
						parameters.corridor_with, rev);
			}
		}
	}

	t2 = 0;
}

void post_proc_Double(float * results_cpu, Reference & ref,
		float *& results_gpu, Parameters parameters, int & t2, long mem_result,
		vector<HelpStruct> helper2, bool rev,
		vector<vector<vector<Match> > > & candidates, bool direction) {

	//copy results form GPU
	CUDA_SAFE_CALL(cudaMemcpy(results_cpu, results_gpu, mem_result
			* sizeof(float), cudaMemcpyDeviceToHost));

	//runs over the results
	for (size_t k = 0; k < t2; k++) {
		for (size_t i = 0; i < candidates[helper2[k].id].size(); i++) { //run over all score functions per submission
			if (parameters.X_best_hits == 1) {
				if (results_cpu[i + k * parameters.match.size() * 2]
						> candidates[helper2[k].id][i][0].score) {

					if (candidates[helper2[k].id][i].size() > 1) {
						candidates[helper2[k].id][i].erase(
								(candidates[helper2[k].id][i].begin() + 1),
								candidates[helper2[k].id][i].end());
					}
					candidates[helper2[k].id][i][0].position
							= helper2[k].position + ref.get_scaff_pos(
									helper2[k].scaff_id); //scaffold pos
					candidates[helper2[k].id][i][0].score = results_cpu[i + k
							* parameters.match.size() * 2]; // score
					candidates[helper2[k].id][i][0].scaffID
							= helper2[k].scaff_id;
					candidates[helper2[k].id][i][0].forward = direction;
					candidates[helper2[k].id][i][0].BS_rev = rev;

				} else if (results_cpu[i + k * parameters.match.size() * 2]
						== candidates[helper2[k].id][i][0].score) {
					int position = helper2[k].position + ref.get_scaff_pos(
							helper2[k].scaff_id);
					int shift = (int)round((double) parameters.corridor_with / 2.0);

					bool flag = true;

					for (vector<Match>::iterator j =
							candidates[helper2[k].id][i].begin(); j
							!= candidates[helper2[k].id][i].end(); j++) {
						if (abs(position - (*j).position) <= shift
								&& (*j).forward == direction && (*j).BS_rev
								== rev && (*j).scaffID == helper2[k].scaff_id) {
							flag = false;
							break;
						}
					}

					if (flag) {
						Match temp;
						temp.position = position; //scaffold pos
						temp.score = results_cpu[i + k
								* parameters.match.size() * 2]; // score
						temp.scaffID = helper2[k].scaff_id;
						temp.forward = direction;
						temp.BS_rev = rev;
						candidates[helper2[k].id][i].push_back(temp);
					}
				}
				//			}
			} else {
				cout << "curr not available " << endl;
				exit(0);
			}
		}
	}

	t2 = 0;
}

//==========================================================================================
//==========================================================================================
//============================== optimal ===================================================
//==========================================================================================
//==========================================================================================

size_t pre_filter_SW_optimal(Reads reads, char *&reads_cpu,
		iterators_str & iter, vector<int> & helper) {
	int t1 = 0;

	for (int i = 0; i < iter.proc; i++) {
		if (iter.readID == iter.numofSeq) {
			//fill up the rest
			cout << "fill" << endl;
			memset(&reads_cpu[i * iter.seqlength], ' ', (iter.proc - i)
					* iter.seqlength * sizeof(char));
			break;
		} else {
			memcpy(&reads_cpu[i * iter.seqlength], (*iter.read_pos).c_str(),
					iter.seqlength * sizeof(char));
		}

		helper[i] = iter.readID;
		iter.read_pos++;
		iter.readID++;
		t1++;
		//		cout<<i<<" "<<iter.readID<<" "<<iter.numofSeq<<endl;
	}
	return t1;

}

void post_proc_SW_optimal(float * score_gpu, float *& score_cpu,
		int * position_gpu, int *& position_cpu, Reads & reads,
		Reference & ref, size_t & t1, long mem_score, long mem_position,
		size_t scaff_id, vector<int> & helper, iterators_str iter, bool rev) {
	cout << "Start1" << endl;
	//copy results form GPU
	CUDA_SAFE_CALL(cudaMemcpy(position_cpu, position_gpu, mem_position
			* sizeof(float), cudaMemcpyDeviceToHost));
	CUDA_SAFE_CALL(cudaMemcpy(score_cpu, score_gpu, mem_score * sizeof(float),
			cudaMemcpyDeviceToHost));

	checkCUDAError("after back copy ");
	cout << "Start2" << endl;
	//runs over the results
	for (size_t k = 0; k < iter.proc; k++) {
		if (helper[k] > 0) {
			if (score_cpu[k] >= reads.matches[helper[k]].match[0].score) {
				reads.matches[helper[k]].match[0].score = score_cpu[k];
				reads.matches[helper[k]].match[0].position = position_cpu[k]
						- iter.size;
				reads.matches[helper[k]].match[0].scaffID = ref.get_scaff_pos(
						scaff_id);
				reads.matches[helper[k]].match[0].BS_rev = rev;
			}
		}
		//
		//		if(score_cpu[k]>4){
		//			cout<<"HIT "<<score_cpu[k]<<" "<<helper[k]<<endl;
		//		}
		helper[k] = -1;
	}
	t1 = 0;
}

//==========================================================================================
//==========================================================================================
//============================== automatically choose parameter ============================
//==========================================================================================
//==========================================================================================

// In preparation (each read gets its own parameters
//=============================================================================================
int call_parameter(int seqlengt) {
	cout << "DO NOT USE JET!!" << endl;
	if (seqlengt < 100) {
		return 2;
	} else {
		return 3;
	}
}
int pre_filter_SW_Auto(map<int, list<ids_str> > &liste, iterators_str & iter,
		char * & reads_cpu, char * & scaff_cpu, vector<HelpStruct> & helper1,
		Reference & scaffold, Parameters parameters, bool & finished) {

	size_t t1 = 0;
	//	int shift= round((double)(parameters.corridor_with)/2.0);
	//	int tresh = round((double)(parameters.corridor_with/3.0));

	if (!finished) {
		for (map<int, list<ids_str> >::iterator i = liste.begin(); i
				!= liste.end(); i++) {
			(*i).second.sort(compare_str);
			(*i).second.unique(is_near_str(iter.tresh));

			for (list<ids_str>::iterator it = (*i).second.begin(); it
					!= (*i).second.end(); ++it) {
				if ((*it).frequ >= parameters.seedword_hits) {
					memcpy(&reads_cpu[t1 * iter.seqlength],
							(*iter.read_pos).c_str(), iter.seqlength
									* sizeof(char));
					int pos = (*it).position - iter.shift;
					if (pos >= 0 && pos + iter.size
							< scaffold.sequence[(*it).scaff_id
									- scaffold.get_ID()].size()) {
						memcpy(
								&scaff_cpu[t1 * iter.size],
								scaffold.sequence[(*it).scaff_id
										- scaffold.get_ID()].substr(
										(size_t) pos, (size_t) iter.size).c_str(),
								iter.size * sizeof(char));
					} else {
						size_t z = 0;
						for (int i = pos; i < pos + iter.size; i++) {
							if (i < 0 || i >= scaffold.sequence[(*it).scaff_id
									- scaffold.get_ID()].size()) {
								scaff_cpu[z + iter.size * t1] = 'N';
							} else {
								scaff_cpu[z + iter.size * t1]
										= scaffold.sequence[(*it).scaff_id
												- scaffold.get_ID()][i];
							}
							z++;
						}
					}
					helper1[t1].id = iter.readID;
					helper1[t1].position = (*it).position;
					if (iter.readID == 11) {
						cout << "Sed ID: 11 pos: " << (*it).position << endl;
					}
					helper1[t1].scaff_id = (*it).scaff_id;
					t1++;
					if (t1 == iter.proc) {
						(*i).second.erase((*i).second.begin(), it);
						finished = false;
						return t1;
					}
				}
			}
			(*i).second.clear();
		}
		finished = true;
		liste.clear();
		iter.readID++;
		iter.mapps++;
		iter.match++;
		iter.read_pos++;
		iter.seed_tresh = call_parameter((*iter.read_pos).size());
	}

	while (iter.readID < iter.numofSeq) {
		if (parameters.max_equal_scoring>1
				|| (*iter.match).match[parameters.X_best_hits - 1].score
						< (*iter.match).orig_length * parameters.match[0]) {
			ids_str temp;
			temp.frequ = 1;
			//collect positions
			for (vector<HashKey>::iterator i = (*iter.mapps).begin(); i
					!= (*iter.mapps).end(); i++) {
				for (vector<pair<int, int> >::iterator j =
						scaffold.lookup_scaffold[(*i).key].begin(); j
						!= scaffold.lookup_scaffold[(*i).key].end(); j++) {
					temp.position = (*j).first - (*i).pos;
					temp.scaff_id = (*j).second;
					liste[(*j).second].push_back(temp);
				}
			}

			for (map<int, list<ids_str> >::iterator j = liste.begin(); j
					!= liste.end(); j++) {
				//filter and count frequencies:
				(*j).second.sort(compare_str);
				(*j).second.unique(is_near_str(iter.tresh));
				for (list<ids_str>::iterator it = (*j).second.begin(); it
						!= (*j).second.end(); ++it) {
					if ((*it).frequ >= iter.seed_tresh) {
						//sum up all positions passed criteria
						memcpy(&reads_cpu[t1 * iter.seqlength],
								(*iter.read_pos).c_str(), iter.seqlength
										* sizeof(char));
						int pos = (*it).position - iter.shift;
						if (pos >= 0 && pos + iter.size
								< scaffold.sequence[(*it).scaff_id
										- scaffold.get_ID()].size()) {
							memcpy(
									&scaff_cpu[t1 * iter.size],
									scaffold.sequence[(*it).scaff_id
											- scaffold.get_ID()].substr(
											(size_t) pos, (size_t) iter.size).c_str(),
									iter.size * sizeof(char));
						} else {
							size_t z = 0;
							for (int i = pos; i < pos + iter.size; i++) {
								if (i < 0 || i
										>= scaffold.sequence[(*it).scaff_id
												- scaffold.get_ID()].size()) {
									scaff_cpu[z + iter.size * t1] = 'N';
								} else {
									scaff_cpu[z + iter.size * t1]
											= scaffold.sequence[(*it).scaff_id
													- scaffold.get_ID()][i];
								}
								z++;
							}
						}
						helper1[t1].id = iter.readID;
						helper1[t1].position = (*it).position;
						if (iter.readID == 11) {
							cout << "Sed ID: 11 pos: " << (*it).position
									<< endl;
						}
						helper1[t1].scaff_id = (*it).scaff_id;
						t1++;
						if (t1 == iter.proc) {
							(*j).second.erase((*j).second.begin(), it);
							finished = false;
							return t1;

						}
					}
				}
				(*j).second.clear();
			}
			liste.clear();
		}
		finished = true;
		iter.readID++;
		iter.mapps++;
		iter.match++;
		iter.read_pos++;
		iter.seed_tresh = call_parameter((*iter.read_pos).size());
	}

	memset(&reads_cpu[t1 * iter.seqlength], ' ', (iter.proc - t1)
			* iter.seqlength * sizeof(char));
	return t1;
}

//==============================================================================================
//==============================================================================================
//==============================================================================================
//=========================PAIR END MODE: ======================================================
//==============================================================================================
//==============================================================================================
//==============================================================================================
//==============================================================================================

bool check_pos(bool direction, int read_pos, int old_pos){//,string old_scaff,string new_scaff) {
//	if(strcmp(old_scaff.c_str(),new_scaff.c_str())==0){
		if (direction) {
			return read_pos > old_pos; //if read is + strand the second part must lie upwards
		}
		return read_pos < old_pos;//if read is + strand the second part must lie downwards;
//	}else{
//		return false;
//	}
}

void pre_filter_SW_pair(long & t1, list<pair <int, int> > & liste,
		iterators_pair_str iter, vector<HelpStruct_pair> & helper1,
		char * & reads_cpu, char * & scaff_cpu, Reference & scaffold,
		Parameters parameters) {


	if (t1 >= iter.proc) {
		cout << "ERROR t1 == proc ?? " << iter.read_ID << endl;
		exit(0);
	}

	pair<int, int> temp;
	temp.second = 1;
	for (vector<HashKey>::iterator i = iter.key_begin; i != iter.key_end; i++) {//get all positions of the key in the reference
		for (vector<pair<int, int> >::iterator j =	scaffold.lookup_scaffold[(*i).key].begin(); j!= scaffold.lookup_scaffold[(*i).key].end(); j++) {

			if ((*j).second == iter.scaffID) {
				temp.first = (*j).first - (*i).pos; //calc theoretically starting position
				if (check_pos((*iter.match).forward, temp.first+ scaffold.get_scaff_pos(iter.scaffID),(*iter.match).position)) {
					liste.push_back(temp);
				}
			}
		}
	}
	liste.sort(compare_pair_str);
	liste.unique(is_near_pair_str(iter.tresh));


	for (list<pair<int, int> >::iterator it = liste.begin(); it != liste.end(); ++it) {

		if ((*it).second >= parameters.seedword_hits) { // if frequency of keys within a position is enough
			//copy read sequence

			memcpy(&reads_cpu[t1 * iter.seqlength], iter.seq.c_str(),iter.seqlength * sizeof(char));

			//copy reference piece
			int pos = (*it).first - iter.shift;
			if (pos >= 0 && pos + iter.size	< scaffold.sequence[iter.scaffID-scaffold.get_ID()].size()) {
				memcpy(&scaff_cpu[t1 * iter.size],scaffold.sequence[iter.scaffID-scaffold.get_ID()].substr((size_t) pos,(size_t) iter.size).c_str(), iter.size* sizeof(char));
			}else{
				//if read overhang on either the beginning or the end of the reference
				size_t z = 0;
				for (int i = pos; i < pos + (int) iter.size; i++) {
					if (i < 0 || i >= scaffold.sequence[iter.scaffID-scaffold.get_ID()].size()) {
						scaff_cpu[z + iter.size * t1] = 'I';
					} else {
						scaff_cpu[z + iter.size * t1]= scaffold.sequence[iter.scaffID-scaffold.get_ID()][i];
					}
					z++;
				}
			}
			//stores meta informations:
			helper1[t1].id = iter.read_ID;
			helper1[t1].position = (*it).first+ scaffold.get_scaff_pos(iter.scaffID);
			helper1[t1].scaff_id = iter.scaffID;
			helper1[t1].BS_rev = iter.BS_rev;
			helper1[t1].forward = iter.forward;
			t1++;

			if (t1 == iter.proc) { //if we collected enough data return
				liste.erase(liste.begin(), it);
				return;
			}
		}
	}
	liste.clear();

}



void pre_filter_SW_pair(long & t1, list<pair <int, int> > & liste,
		iterators_pair_str iter, vector<HelpStruct_pair> & helper1,
		char * & reads_cpu, char * & scaff_cpu, Reference & scaffold,
		Parameters parameters,vector< vector< int > > bslookup, bool * &bs) {


	if (t1 >= iter.proc) {
		cout << "ERROR t1 == proc ?? " << iter.read_ID << endl;
		exit(0);
	}

	pair<int, int> temp;
	temp.second = 1;
	for (vector<HashKey>::iterator i = iter.key_begin; i != iter.key_end; i++) {//get all positions of the key in the reference
		for (vector<pair<int, int> >::iterator j =	scaffold.lookup_scaffold[(*i).key].begin(); j!= scaffold.lookup_scaffold[(*i).key].end(); j++) {

			if ((*j).second == iter.scaffID) {
				temp.first = (*j).first - (*i).pos; //calc theoretically starting position
				if (check_pos((*iter.match).forward, temp.first+ scaffold.get_scaff_pos(iter.scaffID),(*iter.match).position)) {
					liste.push_back(temp);
				}
				//get all position for every permutation in the reference
				for(vector<int>::iterator k= bslookup[(*i).key].begin(); k!=bslookup[(*i).key].end();k++){
					for (vector<pair<int, int> >::iterator u =	scaffold.lookup_scaffold[(*k)].begin(); u!= scaffold.lookup_scaffold[(*k)].end(); u++) {
						if ((*u).second == iter.scaffID) {
							temp.first = (*u).first - (*i).pos; //calc theoretically starting positon
							if (check_pos((*iter.match).forward, temp.first+ scaffold.get_scaff_pos(iter.scaffID),(*iter.match).position)) {
								liste.push_back(temp);
							}
						}

					}
				}
			}
		}
	}
	liste.sort(compare_pair_str);
	liste.unique(is_near_pair_str(iter.tresh));


	for (list<pair<int, int> >::iterator it = liste.begin(); it != liste.end(); ++it) {

		if ((*it).second >= parameters.seedword_hits) { // if frequency of keys within a position is enough
			//copy read sequence

			memcpy(&reads_cpu[t1 * iter.seqlength], iter.seq.c_str(),iter.seqlength * sizeof(char));

			//copy reference piece
			int pos = (*it).first - iter.shift;
			if (pos >= 0 && pos + iter.size	< scaffold.sequence[iter.scaffID-scaffold.get_ID()].size()) {
				memcpy(&scaff_cpu[t1 * iter.size],scaffold.sequence[iter.scaffID-scaffold.get_ID()].substr((size_t) pos,(size_t) iter.size).c_str(), iter.size* sizeof(char));
			}else{
				//if read overhang on either the beginning or the end of the reference
				size_t z = 0;
				for (int i = pos; i < pos + (int) iter.size; i++) {
					if (i < 0 || i >= scaffold.sequence[iter.scaffID-scaffold.get_ID()].size()) {
						scaff_cpu[z + iter.size * t1] = 'I';
					} else {
						scaff_cpu[z + iter.size * t1]= scaffold.sequence[iter.scaffID-scaffold.get_ID()][i];
					}
					z++;
				}
			}

			//stores meta informations:
			helper1[t1].id = iter.read_ID;
			helper1[t1].position = (*it).first+ scaffold.get_scaff_pos(iter.scaffID);
			helper1[t1].scaff_id = iter.scaffID;
			helper1[t1].BS_rev = iter.BS_rev;
			helper1[t1].forward = iter.forward;
			bs[t1]=iter.BS_rev;

			t1++;

			if (t1 == iter.proc) { //if we collected enough data return
				liste.erase(liste.begin(), it);
				return;
			}
		}
	}
	liste.clear();
}


void pre_filter_SW_pair_rest(long & t1,  list< pair<int, int> >  & liste,
		iterators_pair_str iter, vector<HelpStruct_pair> & helper1,
		char * & reads_cpu, char * & scaff_cpu, Reference & scaffold,
		Parameters parameters) {
	//copy regions + reads
	for (list<pair<int, int> >::iterator it = liste.begin(); it != liste.end(); ++it) {
		if ((*it).second >= parameters.seedword_hits) { // if frequency of keys within a position is enough
			//copy read sequence
			memcpy(&reads_cpu[t1 * iter.seqlength], iter.seq.c_str(),iter.seqlength * sizeof(char));
			//copy reference piece
			int pos = (*it).first - iter.shift;

			if (pos >= 0 && pos + iter.size	< scaffold.sequence[iter.scaffID-scaffold.get_ID()].size()) {
				memcpy(&scaff_cpu[t1 * iter.size],scaffold.sequence[iter.scaffID-scaffold.get_ID()].substr((size_t) pos,(size_t) iter.size).c_str(), iter.size* sizeof(char));
			} else {
				//if read overhang on either the begining or the end of the reference
				size_t z = 0;
				for (int i = pos; i < pos + iter.size; i++) {
					if (i < 0 || i >= scaffold.sequence[iter.scaffID-scaffold.get_ID()].size()) {
						scaff_cpu[z + iter.size * t1] = 'N';
					} else {
						scaff_cpu[z + iter.size * t1]= scaffold.sequence[iter.scaffID-scaffold.get_ID()][i];
					}
					z++;
				}
			}

			//stores meta informations:
			helper1[t1].id = iter.read_ID;
			helper1[t1].position = (*it).first+ scaffold.get_scaff_pos(iter.scaffID);
			helper1[t1].scaff_id = iter.scaffID;
			helper1[t1].BS_rev = iter.BS_rev;
			helper1[t1].forward = iter.forward;
			t1++;

			if (t1 == iter.proc) { //if we collected enough data return
				liste.erase(liste.begin(), it);
				return;
			}
		}
	}

	liste.clear();
}


void pre_filter_SW_pair_rest_BS(long & t1,  list< pair<int, int> >  & liste,
		iterators_pair_str iter, vector<HelpStruct_pair> & helper1,
		char * & reads_cpu, char * & scaff_cpu, Reference & scaffold,
		Parameters parameters, bool * & bs_cpu) {
	//copy regions + reads
	for (list<pair<int, int> >::iterator it = liste.begin(); it != liste.end(); ++it) {
		if ((*it).second >= parameters.seedword_hits) { // if frequency of keys within a position is enough
			//copy read sequence
			memcpy(&reads_cpu[t1 * iter.seqlength], iter.seq.c_str(),iter.seqlength * sizeof(char));
			//copy reference piece
			int pos = (*it).first - iter.shift;

			if (pos >= 0 && pos + iter.size	< scaffold.sequence[iter.scaffID-scaffold.get_ID()].size()) {
				memcpy(&scaff_cpu[t1 * iter.size],scaffold.sequence[iter.scaffID-scaffold.get_ID()].substr((size_t) pos,(size_t) iter.size).c_str(), iter.size* sizeof(char));
			} else {
				//if read overhang on either the begining or the end of the reference
				size_t z = 0;
				for (int i = pos; i < pos + iter.size; i++) {
					if (i < 0 || i >= scaffold.sequence[iter.scaffID-scaffold.get_ID()].size()) {
						scaff_cpu[z + iter.size * t1] = 'N';
					} else {
						scaff_cpu[z + iter.size * t1]= scaffold.sequence[iter.scaffID-scaffold.get_ID()][i];
					}
					z++;
				}
			}

			//stores meta informations:
			helper1[t1].id = iter.read_ID;
			helper1[t1].position = (*it).first+ scaffold.get_scaff_pos(iter.scaffID);
			helper1[t1].scaff_id = iter.scaffID;
			helper1[t1].BS_rev = iter.BS_rev;
			helper1[t1].forward = iter.forward;
			bs_cpu[t1]=iter.BS_rev;

			t1++;

			if (t1 == iter.proc) { //if we collected enough data return
				liste.erase(liste.begin(), it);
				return;
			}
		}
	}

	liste.clear();
}

void post_proc_pair(vector<HelpStruct_pair> & helper1, Reads &pairs, long & t1,float * results_cpu, float * results_gpu, long mem_result, Reads reads,Reference & scaffold, size_t tresh) {

	CUDA_SAFE_CALL(cudaMemcpy(results_cpu, results_gpu, mem_result* sizeof(float), cudaMemcpyDeviceToHost));
	//runs over the results
	checkCUDAError("post proc 1");
	for (size_t k = 0; k < t1; k++) {
		if (pairs.matches[helper1[k].id].match.size() == 0 || results_cpu[k]> pairs.matches[helper1[k].id].match[0].score) { //TODO: nbest positions

			pairs.matches[helper1[k].id].match.clear();

			Match temp;
			temp.score = results_cpu[k];
			temp.BS_rev = helper1[k].BS_rev;
			temp.scaffID = helper1[k].scaff_id;
			temp.forward = helper1[k].forward;

//			scaffpos + ref.get_scaff_pos(scaff_id);

			//added the scaffold relative position in the pre filtering
			temp.position = helper1[k].position;// + scaffold.get_scaff_pos(helper1[k].scaff_id);
			pairs.matches[helper1[k].id].match.push_back(temp);

		} else if (results_cpu[k]== pairs.matches[helper1[k].id].match[0].score) {

			//check if duplicate and store
			int position = helper1[k].position;// + scaffold.get_scaff_pos(helper1[k].scaff_id);
			int shift =(int) round((double) tresh / 2.0);
			bool flag = true;
			for (vector<Match>::iterator i =pairs.matches[helper1[k].id].match.begin(); i!= pairs.matches[helper1[k].id].match.end(); i++) {
				if (abs(position - (*i).position) <= shift && (*i).forward== helper1[k].forward && (*i).BS_rev== helper1[k].BS_rev && (*i).scaffID== helper1[k].scaff_id) {
					flag = false;
					break;
				}
			}
			if (flag) {
				Match match;
				match.position = position; //scaffold pos
				match.score = results_cpu[k]; // score
				match.forward = helper1[k].forward;
				match.scaffID = helper1[k].scaff_id;
				match.BS_rev = helper1[k].BS_rev;
				pairs.matches[helper1[k].id].match.push_back(match);
			}
		}
	}
	t1 = 0;
}

