/*
 * Parser.cpp
 *
 *  Created on: Oct 28, 2008
 *      Author: fritz
 */
#include <fstream>
#include <cstdlib>
#include <cstring>

#include "reads_parser.h"
#include "helper/helpers.h"

using std::ifstream;
using std::cout;
using std::endl;
using std::atof;
using std::ios;

void check_file(ifstream myfile, string filename) {
	if (!myfile.good() || myfile.fail()) {
		cout << "There is a problem with the file: " << filename << endl;
		program_close(-1);

	}
}
string comp(string seq) {

	string temp;
	for (size_t j = 0; j < seq.size(); j++) {
		switch (seq[j]) {
		case 'A':
			temp += 'T';
			break;
		case 'C':
			temp += 'G';
			break;
		case 'G':
			temp += 'C';
			break;
		case 'T':
			temp += 'A';
			break;
		default:
			temp += 'N';
			break;
		}
	}
	return temp;
}

string reverse_comp(string seq) {
	string temp;
	for(string::reverse_iterator j = seq.rbegin();j!=seq.rend();j++){
//	for (int j = seq.size() - 1; j >= 0; j--) {
		switch ((*j)) {
		case 'A':
			temp += 'T';
			break;
		case 'C':
			temp += 'G';
			break;
		case 'G':
			temp += 'C';
			break;
		case 'T':
			temp += 'A';
			break;
		default:
			temp += 'N';
			break;
		}
	}

	return temp;
}

void store_read(Reads & reads, Matches temp2, string name, string sequence,string qual) {
	//store the previous read

	if (sequence.size() > reads.get_max_length()) { //checks if we have a new maxlength
		reads.set_max_length((int) sequence.size());
	}

	// store the read, plus additional informations
	temp2.orig_length = (int) sequence.size();
	temp2.qualitvals=qual;
	temp2.read_name = name;
	reads.matches.push_back(temp2);
	reads.backward.push_back(reverse_comp(sequence)); //stores the reverse complement of the read
	reads.forward.push_back(sequence);
}

// read in reads
void getPartOfReads(Reads & reads, Parameters parameters) {

	ifstream myfile;

	reads.set_max_length(0);
	int deleted = 0;

	myfile.open(reads.get_filename(), ifstream::in);

	if (!myfile.good() || myfile.fail()) {
		cout << "There is a problem with the file: " << reads.get_filename() << endl;
		program_close(-1);

	}

	char buffer = ' ';

	//checks which file type (fasta or fastq)
	if (reads.get_type() == ' ') {
		buffer = myfile.get();
		if (buffer == '>') { //fasta file type
			reads.set_type(buffer);
		} else if (buffer == '@') { //fastq file type
			reads.set_type(buffer);
		} else {
			cout << "unknown format!" << endl;
			program_close(parameters.GPU_id);
		}
	} else {
		myfile.seekg(reads.get_filepos()); //jumps to previous used position
		myfile.seekg(-1,ios::cur);
	}
	Matches temp2;
	//loop if more then 1 match is needed

	//init match str,
	Match match;
	match.position = -14;

	match.score = -140000000.0f;
	match.identity = -100.0f;
	//TODO need to change that
	for (int i = 0; i < parameters.X_best_hits; i++) {
		temp2.match.push_back(match);
	}
	string sequence;
//	size_t ncount = 0;
	reads.set_file_pos(-1);
	string name;
	string qual="";
	while (myfile.good()) {

		if (buffer == '+') {
			//  read name
			while (myfile.good() && buffer != '\0' && buffer != '\n') {
				buffer = myfile.get();
			}
			buffer = myfile.get();
			// quality values
			qual.clear();

			while (myfile.good() && buffer != '\0' && buffer != '\n') {

				if(parameters.format!=0){
					if(buffer!=' ' && buffer!='\0' && buffer != '\n'){
						qual+=buffer;
					}
				}
				buffer= myfile.get();
			}

		} else if (buffer == reads.get_type()) { //header

//			if (ncount <= parameters.N_allowed && sequence.size() > parameters.wordsize) {
//			if(reads.forward.size()==0){
//				cout<<name<<" "<<sequence<<endl;
//			}
			if (sequence.size() > parameters.wordsize) {
				store_read(reads, temp2, name, sequence,qual);
			}else if(sequence.size() >0){
				deleted++;
			}

			if (reads.forward.size() == parameters.max_reads_per_run) {
				reads.set_file_pos(myfile.tellg());//breaks if the maximal number of reads are reached. Stores the file position
				break;
			}


			//read name
			sequence.clear();
//			ncount = 0;
			name.clear();
			buffer = myfile.get();

			while (myfile.good() && buffer != '\0' && buffer != '\n') { //reads in the name of the read
				if (name.size() > 3 && buffer == ' ') {
					break;
				}
				name += buffer;
				buffer = myfile.get();
			}
			if (buffer == ' ') { // runs until the end of the line (if needed)
				while (myfile.good() && buffer != '\0' && buffer != '\n') {
					buffer = myfile.get();
				}
			}

		} else if (name.size() != 0) { //read sequence
			buffer = toupper(buffer);
			while (myfile.good() && buffer != '\0' && buffer != '\n') { //reads in the sequence of the read
//				if (sequence.size() > 0 || buffer != 'N') {
					sequence += buffer;
//					if (buffer == 'N') {//!= 'A' && buffer != 'C' && buffer != 'G' &&buffer != 'T'){ //detect unresolved positions
//						ncount++;
//					}
//				}
				buffer = toupper(myfile.get()); //converts all chars to uper case (just in case)
			}
		}
		buffer = myfile.get();
	}

	if (!myfile.good()) {
		if ( sequence.size() > parameters.wordsize) {
//			cout<<name<<" "<<sequence<<endl;
			store_read(reads, temp2, name, sequence,qual);
		} else{
			deleted++;
		}
	}
	qual.clear();
	name.clear();
	sequence.clear();
	myfile.close();

	//prints summary:
	cout << '\t' << "Number of Reads: total:" << reads.forward.size() + deleted << " remain: " << reads.forward.size() << " deleted: " << deleted << endl;
	cout << '\t' << "Length: " << reads.get_max_length() << endl;


	if (reads.forward.size() == 0) {
		//should not be thrown!
		cout << "no sequences read in!" << endl;
		program_close(parameters.GPU_id);

	}
}


// read in reads
void getPartOfReads_paired(Reads & reads, vector<Matches> matches, Parameters parameters, size_t max_size) {

	ifstream myfile;

	reads.set_max_length(0);
	int deleted = 0;
	//	int doubles =0;

	myfile.open(reads.get_filename(), ifstream::in);

	if (!myfile.good() || myfile.fail()) {
		cout << "There is a problem with the file: " << reads.get_filename() << endl;
		program_close(-1);

	}

	char buffer = ' ';

	bool flag=false;
	//checks which file type (fasta or fastq)
	if (reads.get_type() == ' ') {
		buffer = myfile.get();
		if (buffer == '>') { //fasta file type
			reads.set_type(buffer);
		} else if (buffer == '@') { //fastq file type
			reads.set_type(buffer);
		} else {
			cout << "unknown format!" << endl;
			program_close(parameters.GPU_id);
		}
	} else {
		myfile.seekg(reads.get_filepos()); //jumps to previous used position
		myfile.seekg(-1,ios::cur);
		flag=true;
	}
	Matches temp2;
	string sequence;
	reads.set_file_pos(-1);
	string name;
	size_t id = 0;
	int tresh=0;
	string qual;
	while(id < matches.size()){

		while (myfile.good()) {

			if (buffer == '+') {
				//  read name
				while (myfile.good() && buffer != '\0' && buffer != '\n') {
					buffer = myfile.get();
				}
				buffer = myfile.get();
				// quality values
				qual.clear();

				while (myfile.good() && buffer != '\0' && buffer != '\n') {

					if(parameters.format!=0){
						if(buffer!=' ' && buffer!='\0' && buffer != '\n'){
							qual+=buffer;
						}
					}
					buffer= myfile.get();
				}
			}else if (buffer == '@'){// reads.get_type()){ //header
				if(sequence.size()>0){

					store_read(reads, temp2, name, sequence,qual);

					reads.set_file_pos(myfile.tellg());//breaks if the maximal number of reads are reached. Stores the file position

					if (reads.forward.size() == max_size) {
						name.clear();
						sequence.clear();
						myfile.close();
						//prints summary:
						cout << '\t' << "Number of Pair:  total:" << reads.forward.size() + deleted << " remain: " << reads.forward.size() << " deleted: " << deleted << endl;
						cout << '\t' << "Lenght: " << reads.get_max_length() << endl;
						return;

					}

				}
				sequence.clear();
				name.clear();
				buffer = myfile.get();

				while (myfile.good() && buffer != '\0' && buffer != '\n') { //reads in the name of the read
					if (name.size() > 3 && buffer == ' ') {
						break;
					}
					name += buffer;
					buffer = myfile.get();
				}
//				//check if name is equal!
				size_t pos = matches[id].read_name.find_last_of("/");
				string substr = matches[id].read_name.substr(0, pos);
				pos = name.find_last_of("/");
				string substr2 = name.substr(0, pos);
//				cout<<name<<endl;

				if (strcmp(substr.c_str(), substr2.c_str()) == 0) {
//					cout<<"hit"<<endl;
					id++;
				} else {
					tresh++;
					if(tresh>100){

						tresh=0;
						sequence.clear();
						for(int t=0;t<10;t++){
							sequence+='N';
						}
						name.clear();
						name="FAKE";
						qual="";

						store_read(reads, temp2, name, sequence,qual);
						sequence.clear();

						myfile.seekg(reads.get_filepos());
						myfile.seekg(-2,ios::cur);
						buffer = myfile.get();
						id++;
					}
					name.clear();
				}

				if (buffer == ' ') { // runs until the end of the line (if needed)
					while (myfile.good() && buffer != '\0' && buffer != '\n') {
						buffer = myfile.get();
					}
				}
				sequence.clear();
			}else if(name.size() > 0){

//				cout<<"seq"<<endl;
				buffer = toupper(buffer);
				while (myfile.good() && buffer != '\0' && buffer != '\n') { //reads in the sequence of the read
					sequence += buffer;
					buffer = toupper(myfile.get()); //converts all chars to uper case (just in case)
				}
			}
			buffer = myfile.get();
		}

		if(myfile.eof() && id<matches.size()){ //second part not found
			tresh++;

			myfile.close();
			myfile.open(reads.get_filename(), ifstream::in);
			myfile.seekg(reads.get_filepos());
			myfile.seekg(-1,ios::cur);
			buffer = myfile.get();
			id++;
		}else if(myfile.eof() && sequence.size()>0){
			if(sequence.size()>0){
				store_read(reads, temp2, name, sequence,qual);

				reads.set_file_pos(myfile.tellg());//breaks if the maximal number of reads are reached. Stores the file position

				if (reads.forward.size() == max_size) {
					name.clear();
					sequence.clear();
					myfile.close();
					//prints summary:
					cout << '\t' << "Number of Pair:  total:" << reads.forward.size() + deleted << " remain: " << reads.forward.size() << " deleted: " << deleted << endl;
					cout << '\t' << "Lenght: " << reads.get_max_length() << endl;
					return;

				}

			}
		}
	}
	qual.clear();
	name.clear();
	sequence.clear();
	myfile.close();
	//prints summary:
	cout << '\t' << "Number of Pair:  total:" << reads.forward.size() + deleted << " remain: " << reads.forward.size() << " deleted: " << deleted << endl;
	cout << '\t' << "Lenght: " << reads.get_max_length() << endl;

	if (reads.forward.size() == 0) {
		//should not be thrown!
		cout << "no sequences read in!" << endl;
		program_close(parameters.GPU_id);

	}
}

