/*
 * Match_Mismatch_CUDA_CPU.cu
 *
 *  Created on: Nov 12, 2009
 *      Author: fritz
 */

#include "../CUDA_CPU.cuh"
#include "Match_CUDA_Functions.cu"
#include "../../CUDA_Kernel/ungaped_alignment/Match_Mismatch_CUDA_Kernel.cu"

void MatchMismatch(Reference &  scaffold,Reads & reads,Parameters parameters){

	long threads = 256;
	long mem_parameters=5;
	long mem_per_thread =  reads.get_max_length()*sizeof(char)+sizeof(float)+reads.get_max_length()*sizeof(char);
	//									reads     					mem_result     scaff
	long mem2 = parameters.memsize-(mem_parameters*sizeof(float));

	int size=reads.get_max_length()+parameters.corridor_with;
	unsigned int blocks=(mem2/(threads*mem_per_thread));
	blocks=blocks-parameters.mps;

	if(blocks>2250){
		blocks=2250;
	}
	int procs=threads*blocks;

	long mem_reads = reads.get_max_length() * procs;
	long mem_scaff= reads.get_max_length()*procs;
	long mem_result = procs;



	char * reads_cpu= new char [mem_reads];
	char * scaff_cpu= new char [mem_scaff];
	float * results_cpu = new float [mem_result];
	float * parameters_cpu = new float[mem_parameters];

	memset(results_cpu,-1.0,mem_result*sizeof(float));

	parameters_cpu[0] =(float)reads.get_max_length();
	parameters_cpu[1] =parameters.match;
	parameters_cpu[2] =parameters.mismatch;
	parameters_cpu[3] =parameters.T_match;
	parameters_cpu[4] =parameters.T_mismatch;


	CUDA_SAFE_CALL(cudaMemcpyToSymbol(param2, parameters_cpu,mem_parameters* sizeof(float)));

	char * scaffold_gpu;
	char * reads_gpu;
	float * results_gpu;

	CUDA_SAFE_CALL(cudaMalloc((void**)&scaffold_gpu,mem_scaff*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&reads_gpu,mem_reads*sizeof(char)));
	CUDA_SAFE_CALL(cudaMalloc((void**)&results_gpu,mem_result*sizeof(float)));


	CUDA_SAFE_CALL(cudaMemset(results_gpu,-1.0,mem_result*sizeof(float)));
	checkCUDAError("after memcpy");

	unsigned int shared_mem = (sizeof(unsigned int))*threads;

	dim3 dimBlock(threads,1);
	dim3 dimGrid(blocks,1); //per matrix 1 thread!

	iterators_str iter;
	iter.readID=0;
//	iter.size=0;
	iter.seqlength=reads.get_max_length();
	iter.match=reads.matches.begin();
	iter.proc =procs;
//	if(parameters.seedword_hits>0){
		iter.seed_tresh=parameters.seedword_hits;
//	}else{
//		iter.seed_tresh=call_parameter(reads.forward[0].size());
//	}

//	iter.shift =0;// round((double)(parameters.corridor_with)/2.0);
	iter.tresh = round((double)(parameters.corridor_with/3.0));

	if(reads.get_direction()){
		iter.mapps=reads.maped_reads.begin();
		iter.read_pos=reads.forward.begin();
		iter.numofSeq=reads.forward.size();
	}else{
		iter.mapps=reads.maped_reads_back.begin();
		iter.read_pos=reads.backward.begin();
		iter.numofSeq=reads.backward.size();
	}

	size_t t1=0;
	size_t t2=0;


	vector<HelpStruct> helper1;
	vector<HelpStruct> helper2;;

	HelpStruct temp;
	temp.id=0;
	temp.position=0;
	temp.scaff_id=0;

	for(size_t i=0;i<=procs;i++){
		helper1.push_back(temp);
		helper2.push_back(temp);
	}

	map< int,list <ids_str > >liste;
	list <ids_str > tmp;
	for(int i=scaffold.get_ID();i<scaffold.name.size();i++){
		liste[i]=tmp;
	}

	long amount=0;
	bool finished=true;

	while(iter.readID<iter.numofSeq){

		t1=pre_filter_Match(liste,iter,reads_cpu,scaff_cpu,helper1,scaffold,parameters,finished);

		//========================================================================
		amount = amount+t1;
		//=============== process result 2: =======================================
		if(t2 > 0){
			post_proc_Match(results_cpu,reads,scaffold,results_gpu, parameters,t2, mem_result,helper2);
		}
		//=======================================================================
		//=========================== call 1 =================================
		if(t1>0){
			CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
			CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));

			if(parameters.mode==0){
				MatchMismatchBS_Cuda_Kernel<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,results_gpu);
			}else{
				MatchMismatch_Cuda_Kernel<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,results_gpu);
			}
		}
		//=====================================================================
		//=================== pre filter 2: ============================================
		if(t1==procs){
			t2 =pre_filter_Match(liste,iter,reads_cpu, scaff_cpu,helper2,scaffold,parameters,finished);
		}

//		========================================================================
		amount = amount+t2;
		//=============== process result 1: =======================================
		if(t1>0){
			post_proc_Match(results_cpu,reads,scaffold,results_gpu, parameters,t1,mem_result,helper1);
		}
		//=======================================================================
		//=========================== call 2 ====================================
		if(t2 > 0){

			CUDA_SAFE_CALL(cudaMemcpy(scaffold_gpu,scaff_cpu,mem_scaff*sizeof(char),cudaMemcpyHostToDevice));
			CUDA_SAFE_CALL(cudaMemcpy(reads_gpu,reads_cpu,mem_reads*sizeof(char),cudaMemcpyHostToDevice));

			if(parameters.mode==0){
				MatchMismatchBS_Cuda_Kernel<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,results_gpu);
			}else{
				MatchMismatch_Cuda_Kernel<<<dimGrid,dimBlock,shared_mem>>>(scaffold_gpu,reads_gpu,results_gpu);
			}
		}
		//=======================================================================
		//=============== process result 2 if it is the end!: =======================================

		//=======================================================================
	}
	if(t2>0){ //Just to be shure that I dont miss the last step!
		post_proc_Match(results_cpu,reads,scaffold,results_gpu, parameters,t2, mem_result,helper2);
	}

	cout<<"\t\t"<<amount<<" where calculated"<<endl;

	delete [] parameters_cpu;
	delete [] results_cpu;
	helper1.clear();
    helper2.clear();
	delete [] reads_cpu;
	delete [] scaff_cpu;


	CUDA_SAFE_CALL(cudaFree(results_gpu));
	CUDA_SAFE_CALL(cudaFree(reads_gpu));
	CUDA_SAFE_CALL(cudaFree(scaffold_gpu));

	checkCUDAError("after delete");
}
